from ctypes import *
from operator import contains
import os
import json

def nr_sdk():    
    cur_path = os.path.dirname(__file__)
    dll_path = cur_path + "/lib/QLNRSdk.dll"
    api = cdll.LoadLibrary(dll_path)
    
    # init
    api.QLNR_Init()
    return api

_api = nr_sdk()

class NRSDK(object):
    def __init__(self) :
        self.ip = r'127.0.0.1'
        self.login_handle = c_void_p(None)
        self.connected = False

    def connect(self, ip=None):
        if ip:
            self.ip = ip
        #login
        server_ip = c_char_p(self.ip.encode('utf-8'))
        res = _api.QLNR_Login(server_ip, pointer(self.login_handle))
        self.connected = (res == 0)

        #log
        print("connect to app server[{}] {}.".format(self.ip, "success" if self.connected else "fail"))

        return self
        
    def close(self):
        print("logout from app server[{}]".format(self.ip))
        _api.QLNR_Logout(self.login_handle)
        self.connected = False
    
    # 未连接时，自动连接一次
    def _assert_connected(self):
        if self.connected:
            return
        
        self.connect()

    # 开始刺激
    def _start_stimulation(self, param, device=None):
        self._assert_connected()

        # 字符串格式统一为dict
        if isinstance(param, str):
            param = json.loads(param)

        s = json.dumps(param, indent=4)
        # print(s)

        param_2 = c_char_p(s.encode('utf-8'))
        p_resp = (c_char * 512)(0)
        len = 512
        device_code = c_char_p(device.encode('utf-8')) if device else c_char_p(None)
        res = -1
        try:
            res = _api.QLNR_StartStimulationEx(self.login_handle, param_2, p_resp, len, device_code)
            # print("start stimulation {}".format(resp_result(res)))
            # if res != _RESP_SUCCESS:
            #     print(p_resp.value.decode("utf-8"))
        except Exception as e:
            print(e)
    
        return _SUCCESS if res == _RESP_SUCCESS else _FAIL
        
    #停止刺激
    def _stop_stimulation(self, device=None):
        self._assert_connected()

        p_resp = (c_char * 512)(0)
        len = 512
        device_code = c_char_p(device.encode('utf-8')) if device else c_char_p(None)
        try:
            res = _api.QLNR_StopStimulationEx(self.login_handle, p_resp, len, device_code)
            # print("stop stimulation {}".format(resp_result(res)))
            # if res != _RESP_SUCCESS:
            #     print(p_resp.value.decode("utf-8"))
        except Exception as e:
            print(e)
        
        return _SUCCESS if res == _RESP_SUCCESS else _FAIL
        
    #获取上位机已连接的设备列表
    def _get_device_list(self):
        self._assert_connected()

        p_resp = (c_char * 512)(0)
        len1 = 512
        p_list = (c_char * 4096)(0)
        len2 = 4096
        res = -1
        try:
            res = _api.QLNR_GetDeviceListEx(self.login_handle, p_resp, len1, p_list, len2)
            # print("get device list {}".format(resp_result(res)))
            if res == _RESP_SUCCESS:
                list = p_list.value.decode("utf-8")
                return json.loads(list)
            else:
                print(p_resp.value.decode("utf-8"))
        except Exception as e:
            print(e)
        
        return []

    #设置wifi参数
    def _set_wifi_config(self, ssid, skey, device=None):
        self._assert_connected()
        print(f"ssid:{ssid}, secret:{skey}: device:{device}")

        p_resp = (c_char * 512)(0)
        len = 512
        res = -1
        p_ssid = c_char_p(ssid.encode('utf-8')) if ssid else c_char_p(None)
        p_skey = c_char_p(skey.encode('utf-8')) if skey else c_char_p(None)
        p_device = c_char_p(device.encode('utf-8')) if device else c_char_p(None)
        try:
            res = _api.QLNR_SetDeviceWifiConfigEx(self.login_handle, p_ssid, p_skey, p_resp, len, p_device)
            # print("set wifi config {} for device[{}]".format(resp_result(res), device))
            if res != _RESP_SUCCESS:
                print(p_resp.value.decode("utf-8"))
        except Exception as e:
            print(e)
        
        return _SUCCESS if res == _RESP_SUCCESS else _FAIL

    #查询wifi参数
    def _get_wifi_config(self, device=None):
        self._assert_connected()

        p_wifi_config = (c_char * 512)(0)
        wifi_config_len = 512
        p_resp = (c_char * 512)(0)
        len = 512
        res = -1
        p_device = c_char_p(device.encode('utf-8')) if device else c_char_p(None)

        result = {}
        try:
            res = _api.QLNR_GetDeviceWifiConfigEx(self.login_handle, p_wifi_config, wifi_config_len, p_resp, len, p_device)
            print("get wifi config {} for device[{}]".format(resp_result(res), device))
            if res == _RESP_SUCCESS:
                wifi_config = p_wifi_config.value.decode("utf-8")
                # print(wifi_config)
                wifi_config = json.loads(wifi_config)
                result["ssid"] = wifi_config["configssid"]
                result["skey"] = wifi_config["configpassword"]
                result["ssid_default"] = wifi_config["defaultssid"]
                result["skey_default"] = wifi_config["defaultpassword"]
                result["ssid_connected"] = wifi_config["currentssid"]
                result["skey_connected"] = wifi_config["currentpassword"]

            else:
                print(p_resp.value.decode("utf-8"))
            
        except Exception as e:
            print(e)
        
        return result

    #重启设备
    def _restart_device(self, device=None):
        self._assert_connected()

        p_resp = (c_char * 512)(0)
        len = 512
        res = -1
        p_device = c_char_p(device.encode('utf-8')) if device else c_char_p(None)

        try:
            res = _api.QLNR_SetDeviceResetEx(self.login_handle, p_resp, len, p_device)
            # print("restart device {} for device[{}], resp: {}".format(resp_result(res), device, p_resp.value.decode("utf-8")))
            
        except Exception as e:
            print(e)
        
        return res

    #事件标记
    def _trigger(self, event_id, device=None):
        self._assert_connected()
        
        p_resp = (c_char * 512)(0)
        len = 512
        res = -1
        p_device = c_char_p(device.encode('utf-8')) if device else c_char_p(None)

        try:
            res = _api.QLNR_SetTrigger(self.login_handle, event_id, p_resp, len, p_device)
            # print("trigger {} for device[{}], resp: {}".format(resp_result(res), device, p_resp.value.decode("utf-8")))
            
        except Exception as e:
            print(e)
        
        return res

    #开启信号采集（并记录到文件）
    def _start_collection(self, param=None, path=None, file_type=None, device=None):
        self._assert_connected()

        p_param = c_char_p(None)
        if param:
            # 字符串格式统一为dict
            if isinstance(param, str):
                param = json.loads(param)

            s = json.dumps(param, indent=4)
            print(s)

            p_param = c_char_p(s.encode('utf-8'))
        p_path = c_char_p(path.encode('utf-8')) if path else c_char_p(None)
        
        p_resp = (c_char * 512)(0)
        len = 512
        res = -1
        p_device = c_char_p(device.encode('utf-8')) if device else c_char_p(None)

        try:
            res = _api.QLNR_StartCollectEx(self.login_handle, p_param, p_path, file_type, p_resp, len, p_device)
            # print("start collection {} for device[{}], resp: {}".format(resp_result(res), device, p_resp.value.decode("utf-8")))
            
        except Exception as e:
            print(e)
        
        return res

    #停止信号采集
    def _stop_collection(self, device=None):
        self._assert_connected()
        
        p_resp = (c_char * 512)(0)
        len = 512
        res = -1
        p_device = c_char_p(device.encode('utf-8')) if device else c_char_p(None)

        try:
            res = _api.QLNR_StopCollectEx(self.login_handle, p_resp, len, p_device)
            # print("start collection {} for device[{}], resp: {}".format(resp_result(res), device, p_resp.value.decode("utf-8")))
            
        except Exception as e:
            print(e)
        
        return res
        

    #获取上位机已连接的设备列表
    def _get_subject_list(self):
        self._assert_connected()

        p_resp = (c_char * 512)(0)
        len1 = 512
        p_list = (c_char * 4096)(0)
        len2 = 4096
        res = -1
        try:
            res = _api.QLNR_GetPatientListEx(self.login_handle, p_resp, len1, p_list, len2)
            # print("get subject list {}".format(resp_result(res)))
            if res == _RESP_SUCCESS:
                list = p_list.value.decode("utf-8")
                return json.loads(list)
            else:
                print(p_resp.value.decode("utf-8"))
        except Exception as e:
            print(e)
        
        return []

    #获取范式列表-和设备相关
    def _get_paradigm_list(self, device=None):
        self._assert_connected()

        p_resp = (c_char * 512)(0)
        len1 = 512
        p_list = (c_char * 40960)(0)
        len2 = 40960
        res = -1
        p_device = c_char_p(device.encode('utf-8')) if device else c_char_p(None)
        try:
            res = _api.QLNR_GetParadigmListEx(self.login_handle, p_resp, len1, p_list, len2, p_device)
            # print("get paradigm list {}".format(resp_result(res)))
            if res == _RESP_SUCCESS:
                list = p_list.value.decode("utf-8")
                return json.loads(list)
            else:
                print(p_resp.value.decode("utf-8"))
        except Exception as e:
            print(e)
        
        return []

    #设置被试
    def _set_subject_select(self, subject_id, device=None):
        self._assert_connected()
        # print(f"_set_subject_select(subject_id:{subject_id}, : device:{device})")

        p_resp = (c_char * 512)(0)
        len = 512
        res = -1
        p_subject_id = c_char_p(subject_id.encode('utf-8')) if subject_id else c_char_p(None)
        p_device = c_char_p(device.encode('utf-8')) if device else c_char_p(None)
        try:
            res = _api.QLNR_SetPatientEx(self.login_handle, p_subject_id, p_resp, len, p_device)
            # print("_set_subject_select {} for device[{}]".format(resp_result(res), device))
            if res != _RESP_SUCCESS:
                print(p_resp.value.decode("utf-8"))
        except Exception as e:
            print(e)
        
        return _SUCCESS if res == _RESP_SUCCESS else _FAIL

    #设置范式
    def _set_paradigm_select(self, paradigm_id, device=None):
        self._assert_connected()
        # print(f"_set_paradigm_select(paradigm_id:{paradigm_id}, : device:{device})")

        p_resp = (c_char * 512)(0)
        len = 512
        res = -1
        p_paradigm_id = c_char_p(paradigm_id.encode('utf-8')) if paradigm_id else c_char_p(None)
        p_device = c_char_p(device.encode('utf-8')) if device else c_char_p(None)
        try:
            res = _api.QLNR_SetParadigmEx(self.login_handle, p_paradigm_id, p_resp, len, p_device)
            # print("_set_paradigm_select {} for device[{}]".format(resp_result(res), device))
            if res != _RESP_SUCCESS:
                print(p_resp.value.decode("utf-8"))
        except Exception as e:
            print(e)
        
        return _SUCCESS if res == _RESP_SUCCESS else _FAIL

_RESP_SUCCESS = 0
_SUCCESS = 1
_FAIL = 0

def resp_result(res):
    return 'success' if res == 0 else 'fail'

def show_result(res):
    return 'success' if res == 1 else 'fail'


from bs4 import BeautifulSoup
from .paradigm import Paradigm
import os

class McfReader():
    def __init__(self, fname, marker_type=None):
        self._fname = fname
        self._blocks = None
        self._marker_type = marker_type
        self._marker_id = None
        self._paradigm = None
        #read
        self._parser()

    @property
    def paradigm(self):
        return self._paradigm

    def _parser(self):
        soup = BeautifulSoup(open(self._fname), 'xml')
        self._blocks = soup.blocks
        # read file name as marker id
        if self._marker_type == 'name':
            self._read_marker()
        self._read_paradigm()

    def _read_marker(self):
        try:
            (fpath, fname) = os.path.split(self._fname)
            self._marker_id = int(fname[:-4])
        except:
            print (f"{self._fname} _read_marker error.")

    def _read_channels(self, blocks):
        rst = dict()
        try:
            electrodes = blocks.find_all("electrode")
            for idx, electrode in enumerate(electrodes):
                ch = electrode["channelID"]
                current = electrode["amplitude"] 
                # current uA to mA
                rst[ch]= float(current) / 1000.0
        except:
            print("_read_channels error.")

        return rst
            
    def _read_paradigm(self):
        try:
            up = self._blocks.find_all(id="0")[0]
            dc = self._blocks.find_all(id="1")[0]
            down = self._blocks.find_all(id="2")[0]
            up_duration = up.duration_ms.text
            duration = dc.duration_ms.text
            down_duration = down.duration_ms.text
            channels = self._read_channels(dc.spatialPattern)

            self._paradigm = Paradigm(channels, up_duration, duration, down_duration, marker_id=self._marker_id)
        except:
            print (f"{self._fname} _read_paradigm error.")




import os
import glob
from datetime import datetime, timezone

TRIGGER_MAPPING = {
    1000: 'start of stimulation',
    1001: 'end of stimulation',
    1201: 'event 1',
    1202: 'event 2',
    1203: 'event 3',
    1204: 'event 4',
    1205: 'event 5',
    1206: 'event 6',
    1207: 'event 7',
    1208: 'event 8',
    1209: 'event 9',
    1210: 'event 10'
}

DEVICE_MAPPING = {
    101: 'A32',
    102: 'A64R',
    103: 'C64',
    104: 'H32S',
    105: 'A32'
}

def bytes_to_number(bs, isBig=False, signed=True):
    if (len(bs) == 3) and not isBig:
        isNeg = bs[0] & 0x80 == 1 and signed
        return (0xFFFFFFFF if isNeg else 0x0) | bs[0]<< 16 | bs[1] << 8  | bs[2]
    return int.from_bytes(bs, byteorder='big' if isBig else 'little', signed=signed)

def bytes_to_datetime(bs, isBig=True):
    year = bytes_to_number(bs[1:3], isBig=True)
    month = bytes_to_number(bs[3:4])
    day = bytes_to_number(bs[4:5])
    hour = bytes_to_number(bs[5:6])
    minute = bytes_to_number(bs[6:7])
    sec = bytes_to_number(bs[7:8])
    ms = bytes_to_number(bs[8:], isBig=True)
    return datetime(year, month, day, hour, minute, sec, ms,
                                 tzinfo=timezone.utc)

        
def get_file_list(fpath, ext='edf'):    
    file_filter = "*.{}".format(ext)
    return glob.glob(os.path.join(fpath, file_filter))


from setuptools import setup
from setuptools import find_packages

VERSION = '1.1.3.4'
AUTHOR='eegion'
EMAIL='hehuajun@eegion.com'
REQUIRED = [
    'beautifulsoup',
    "lxml"
]

option = {
    "build_exe": {
        "packages":[
            'beautifulsoup',
            "lxml"
        ],
        "excludes":["test", "main"]
    }
}

setup(
    name='nrsdk',  # package name
    version=VERSION,  # package version
    author=AUTHOR,
    author_email=EMAIL,
    requires=REQUIRED,
    description='Api for use quanlan device, since v1.1.3',  # package description
    packages=find_packages(),
    package_data={
        "nrsdk": ["lib/*.dll"],
        "":["*.txt", "*.md"]
    },
    zip_safe=False
)


from __future__ import annotations # for Python 3.7-3.9
import enum
import json
from typing import TypedDict
from functools import wraps

def to_list(in_ch):
    ret = []
    for ch in in_ch:
        assert ch % 1 == 0, 'ch must be int'
        ret.append(int(ch))
    return ret

def to_int(in_int):
    assert in_int % 1 == 0, 'ch must be int'
    return int(in_int)

class WaveForm(enum.IntEnum):
    DC = 0,
    AC = 1,
    BidPulse = 3 

class NrParamChannel(TypedDict):
    waveform: Union[WaveForm,int]
    frequency:int = 0
    current: Union[int,float]
    ramp_up:Union[int,float] = 0
    ramp_down:Union[int,float] = 0
    duration:Union[int,float]
    phase_position:Union[int,float] = 0
    channel_id:int = 0

    @classmethod
    def dc(cls, current:float, 
            *, channel_id = 0, duration = 0xFFFFFFFF, ramp_up = 0, ramp_down = 0)->NrParamChannel:
        return {
                "waveform":WaveForm.DC,
                "current": current,
                "duration":duration,
                "ramp_up":ramp_up,
                "ramp_down":ramp_down,
                "channel_id":channel_id,
                }

    
    @classmethod
    def ac(cls, current:float, frequency:int,
            *, duration=0xFFFFFFFF, ramp_up = 0, ramp_down = 0, phase_position = 0
            ) ->NrParamChannel:
        return {
            "waveform":WaveForm.AC,
            "current": current,
            "duration":duration,
            "ramp_up":ramp_up,
            "ramp_down":ramp_down,
            "channel_id":0,
            "phase_position":phase_position,
            }


class NrParamParams(TypedDict):
    channels: list[NrParamChannel]
    @classmethod
    def default(cls)->dict:
        return {"channels": []}

class NrParamBase(TypedDict):
    params:List[NrParamParams]
    @classmethod
    def default(cls)->dict:
        return {"params": [NrParamParams.default()]}

class NrParam:
    @classmethod
    @wraps(NrParamChannel.ac)
    def ac(cls, *arg, **kwargs)->NrParamChannel:
        return NrParamChannel.ac(*arg, **kwargs)

    @classmethod
    @wraps(NrParamChannel.dc)
    def dc(cls, *arg, **kwargs)->NrParamChannel:
        return NrParamChannel.dc(*arg, **kwargs)

    def get_param(self) -> NrParamBase:
        return self.param

    # ch_id_start 决定channel_apply时通道id从0或1 开始
    def __init__(self, ch_id_start = 0):
        self.ch_id_start = ch_id_start
        self.param:NrParamBase = NrParamBase.default()
    
    def channel_apply(self, channel_id:Union[list, int], wave:NrParamChannel, *, 
                    channel_id_nega:Union[list, int] = [], duration = None, ramp_up = None, ramp_down = None)->None:

        channel_list:list[NrParamChannel] = self.param["params"][0]["channels"]
        def app_int(ch_id:int, scale:float):
            ch_id = to_int(ch_id) 
            assert ch_id >= self.ch_id_start
            ch_wave = wave.copy()
            ch_wave['channel_id'] = ch_id - self.ch_id_start # TODO
            if duration is not None:
                ch_wave['duration'] = duration
            if ramp_up is not None:
                ch_wave['ramp_up'] = ramp_up
            if ramp_down is not None:
                ch_wave['ramp_down'] = ramp_down
            ch_wave['current'] *= scale

            channel_list.append(ch_wave)

        try:
            int(channel_id)
        except:
            pass
        else:
            channel_id = [channel_id]
        
        try:
            int(channel_id_nega)
        except:
            pass
        else:
            channel_id_nega = [channel_id_nega]
        
        scals = {}
        if len(channel_id) == 0:
            for i in channel_id_nega:
                scals[i] = -1
        elif len(channel_id_nega) != 0:
            posi_s = 1/len(channel_id)
            posi_n = -1/len(channel_id_nega)
            for i in channel_id_nega:
                scals[i] = posi_n
            for i in channel_id:
                scals[i] = posi_s
        channel_id = list(set(channel_id)|set(channel_id_nega))
        for i in channel_id:
            app_int(i, scals.get(i, 1))

if __name__ == '__main__':      
    p = NrParam() # 
    dc = p.dc(1)
    ac = p.ac(1, 10)
    p.channel_apply([4,3,5], ac, channel_id_nega=[2], duration=100)
    pr = json.dumps(p.get_param(), indent=2)
    print(pr)
